package com.bluetoothconnection.billy.bluetoothconnection;

import android.Manifest;
import android.annotation.TargetApi;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageManager;
import android.media.AudioManager;
import android.os.Handler;
import android.os.Message;
import android.provider.ContactsContract;
import android.support.v4.app.ActivityCompat;
import android.support.v7.app.AppCompatActivity;
import android.os.Bundle;
import android.view.View;
import android.widget.AdapterView;
import android.widget.ArrayAdapter;
import android.widget.ListView;
import android.widget.TextView;
import android.widget.Toast;
import android.util.Log;
import android.net.Uri;
import android.telecom.Call;

import org.w3c.dom.Text;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Set;
import java.util.UUID;

// public class MainActivity extends AppCompatActivity
public class MainActivity extends AppCompatActivity implements AdapterView.OnItemClickListener{

    public boolean isConnected = false;
    public String currentState = STATE_FIRST_START;
    public boolean socketConnected = false;
    public int readFail = 0;

    ArrayAdapter<String> listAdapter;
    ListView listView;
    BluetoothAdapter btAdapter;
    Set<BluetoothDevice> devicesArray;
    ArrayList<String> pairedDevices;
    ArrayList<BluetoothDevice> devices;
    public static final UUID MY_UUID = UUID.fromString("00001101-0000-1000-8000-00805F9B34FB");  // UUID for most bluetooth devices "00001101-0000-1000-8000-00805F9B34FB"
    protected static final int SUCCESS_CONNECT = 0;
    protected static final int MESSAGE_READ = 1;
    protected static final int MESSAGE_CONNECT_FAILED = 2;
    protected static final String STATE_FIRST_START = "10";
    protected static final String STATE_CALL_MADE = "11";
    protected static final String STATE_CALL_ENDED = "12";
    // UUID from UUID generator        "7d74b9c8-7e85-4485-9863-deca8a4ccb4b"
    IntentFilter filter;
    BroadcastReceiver receiver;

    //***************************************************************************************************//
    //Handler, was originally supposed to be used from within the two bluetooth threads, but was having
    //issues so switched to a method call, I don't believe this is actually being used in the code.
    //***************************************************************************************************//

    Handler mHandler = new Handler(){
        @Override
        public void handleMessage(Message msg) {
            // TODO Auto_generated method stub
            super.handleMessage(msg);
            switch(msg.what){
                case SUCCESS_CONNECT:
                    ConnectedThread connectedThread = new ConnectedThread((BluetoothSocket)msg.obj);
                    Toast.makeText(getApplicationContext(), "Connected", Toast.LENGTH_SHORT);
                    String s = " successfully connected";
                    connectedThread.write(s.getBytes());
                    break;
                case MESSAGE_READ:
                    byte[] readBuf = (byte[])msg.obj;
                    String ss = new String(readBuf);
                    Toast.makeText(getApplicationContext(), ss, Toast.LENGTH_SHORT);
                    Log.e("TAG", ss);
                    break;
                case MESSAGE_CONNECT_FAILED:
                    TextView textView = (TextView)findViewById(R.id.tvConnected);
                    textView.setText((String)"Socket failed...try again");
                    break;
            }
            //TextView textView = (TextView)findViewById(R.id.tvConnected);
            //textView.setText((String)"Socket failed...try again");
        }
    };

    //***************************************************************************************************//
    //Initializes the needed variables, checks to make sure bluetooth is on and working, gets bluetooth
    //devices (paired and unpaired)
    //***************************************************************************************************//
    @Override
        protected void onCreate(Bundle savedInstanceState) {
            if (!isConnected)
            {
                super.onCreate(savedInstanceState);
                //currentState = STATE_FIRST_START;
                init();
                if(btAdapter==null){
                    Toast.makeText(getApplicationContext(), "No bluetooth detected", Toast.LENGTH_SHORT).show();
                    finish();
                }
                else{
                    if(!btAdapter.isEnabled()){
                        turnOnBT();
                    }

                    getPairedDevices();
                    startDiscovery();
                }

                mHandler = new Handler();
            }

            //***************************************************************************************************//
            //Trying to find when a call ends...no luck yet, most likely caused by this executing before the
            //thread changed the currentState value.
            //***************************************************************************************************//
            else if (isConnected)
            {
                ListView listView = (ListView)findViewById(R.id.listView);
                listView.clearChoices();
                listView.setClickable(false);
                TextView textView = (TextView)findViewById(R.id.tvConnected);
                textView.setText("Connected");
                currentState = STATE_CALL_ENDED;

            }

    }

    //***************************************************************************************************//
    //Starts discovering bluetooth devices, preemptively cancels discovery incase it was already running
    //***************************************************************************************************//
    private void startDiscovery() {
        btAdapter.cancelDiscovery();
        btAdapter.startDiscovery();
    }

    //***************************************************************************************************//
    //Creates the request to turn on bluetooth on the android device if the bluetooth is not already on
    //***************************************************************************************************//
    private void turnOnBT() {
        Intent intent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
        startActivityForResult(intent, 1);
    }

    //***************************************************************************************************//
    //Gets all devices currently paired with this device and puts them into an array
    //***************************************************************************************************//
    private void getPairedDevices() {
        // TODO Auto-generated method stub
        devicesArray = btAdapter.getBondedDevices();
        if(devicesArray.size()>0){
            for(BluetoothDevice device:devicesArray){
                pairedDevices.add(device.getName());
                //pairedDevices.add(device.getName()+"\n"+device.getAddress());

            }
        }
    }

    //***************************************************************************************************//
    //Initializes values to be used throughout the rest of the program, creates the onReceive event
    //that detects nearby bluetooth devices and adds them to the ListView
    //***************************************************************************************************//
    private void init() {
        // TODO Auto-generate method stub
        setContentView(R.layout.activity_main);
        listView = (ListView)findViewById(R.id.listView); // =(ListView)findView.....
        listView.setOnItemClickListener(this);
        listAdapter= new ArrayAdapter<>(this, android.R.layout.simple_list_item_1, 0);
        listView.setAdapter(listAdapter);
        btAdapter = BluetoothAdapter.getDefaultAdapter();
        pairedDevices = new ArrayList<>();
        filter = new IntentFilter(BluetoothDevice.ACTION_FOUND);
        devices = new ArrayList<>();
        receiver = new BroadcastReceiver() {
            @Override
            public void onReceive(Context context, Intent intent) {
                // TODO Auto-generated method stub
                String action = intent.getAction();

                if(BluetoothDevice.ACTION_FOUND.equals(action)){
                    BluetoothDevice device = intent.getParcelableExtra(BluetoothDevice.EXTRA_DEVICE);
                    devices.add(device);

                    String s = "";
                    for(int i=0; i < pairedDevices.size(); i++){

                        if (device.getName().equals(pairedDevices.get(i))) {
                            s = "(Paired)";
                            break;
                        }

                    }
                    listAdapter.add(device.getName()+" "+s+" "+"\n"+device.getAddress());
                }
                else if(BluetoothAdapter.ACTION_DISCOVERY_STARTED.equals(action)){
                    Toast.makeText(context, "Discovery Started", Toast.LENGTH_SHORT).show();
                }
                else if(BluetoothAdapter.ACTION_DISCOVERY_FINISHED.equals(action)){
                    Toast.makeText(context, "Discovery Finished", Toast.LENGTH_SHORT).show();
                }
                else if(BluetoothAdapter.ACTION_STATE_CHANGED.equals(action)){
                    if(btAdapter.getState() == btAdapter.STATE_OFF){
                        turnOnBT();
                    }
                }
            }
        };


        registerReceiver(receiver, filter);
        filter = new IntentFilter(BluetoothAdapter.ACTION_DISCOVERY_STARTED);
        registerReceiver(receiver, filter);
        filter = new IntentFilter(BluetoothAdapter.ACTION_DISCOVERY_FINISHED);
        registerReceiver(receiver, filter);
        filter = new IntentFilter(BluetoothAdapter.ACTION_STATE_CHANGED);
        registerReceiver(receiver, filter);

    }

    //***************************************************************************************************//
    //Event occurs when application pauses
    //***************************************************************************************************//
    @Override
    protected void onPause() {
        // TODO Auto-generated method stub
        super.onPause();
        //unregisterReceiver(receiver);
    }

    //***************************************************************************************************//
    //Occurs when app resumes, attempted using this to identify when call ends
    //***************************************************************************************************//
    @Override
    public void onResume(){
        super.onResume();
        if (isConnected)
        {
            currentState = STATE_CALL_ENDED;
            setContentView(R.layout.activity_main);
            ListView listView = (ListView)findViewById(R.id.listView);
            listView.clearChoices();
            listView.setClickable(false);
            TextView textView = (TextView)findViewById(R.id.tvConnected);
            if(socketConnected)
            {
                textView.setText((String)"Connected");
            }
            else
            {
                textView.setText((String)"Socket failed...try again");
            }

        }
    }

    //***************************************************************************************************//
    //Checks if bluetooth gets canceled
    //***************************************************************************************************//
    @Override
    protected void onActivityResult(int requestCode, int resultCode, Intent data) {
        // TODO Auto-generated method stub
        super.onActivityResult(requestCode, resultCode, data);
        if(resultCode == RESULT_CANCELED){
            Toast.makeText(getApplicationContext(), "Bluetooth must be enabled to continue", Toast.LENGTH_SHORT).show();
            finish();
        }
    }

    //***************************************************************************************************//
    //Start by canceling discover (discovery must be stopped to proceed with the program)
    //Checks what bluetooth device from the list was clicked, if that device is paired then
    //a connection will attempt to be made.  If not paired, it informs the user and does nothing.
    //***************************************************************************************************//
    @Override
    public void onItemClick(AdapterView<?> adapterView, View view, int i, long l) {
        // TODO Auto-generated method stub
        if(btAdapter.isDiscovering()){
            btAdapter.cancelDiscovery();
        }
        if(listAdapter.getItem(i).contains("Paired")){
            Toast.makeText(getApplicationContext(), "device is paired", Toast.LENGTH_SHORT).show();
            BluetoothDevice selectedDevice = devices.get(i);
            ConnectThread connect = new ConnectThread(selectedDevice);
            connect.start();
            //connect.run();

            setContentView(R.layout.activity_main);
            ListView listView = (ListView)findViewById(R.id.listView);
            listView.clearChoices();
            listView.setClickable(false);
            TextView textView = (TextView)findViewById(R.id.tvConnected);
            textView.setText((String)"Connected");


        }
        else{
            Toast.makeText(getApplicationContext(), "device is not paired", Toast.LENGTH_SHORT).show();
        }
    }

    //Bluetooth connect thread from developer.android.com
    private class ConnectThread extends Thread {
        private final BluetoothSocket mmSocket;
        private final BluetoothDevice mmDevice;

        public ConnectThread(BluetoothDevice device) {
            // Use a temporary object that is later assigned to mmSocket
            // because mmSocket is final.
            BluetoothSocket tmp = null;
            mmDevice = device;

            try {
                // Get a BluetoothSocket to connect with the given BluetoothDevice.
                // MY_UUID is the app's UUID string, also used in the server code.
                tmp = device.createRfcommSocketToServiceRecord(MY_UUID);
            } catch (IOException e) {}
            mmSocket = tmp;
        }

        public void run() {
            // Cancel discovery because it otherwise slows down the connection.
            btAdapter.cancelDiscovery();

            try {
                // Connect to the remote device through the socket. This call blocks
                // until it succeeds or throws an exception.
                mmSocket.connect();
            } catch (IOException connectException) {
                // Unable to connect; close the socket and return.
                Log.e("TAG", "Socket didn't connect");
                try {
                    Log.e("TAG", "Closing Socket");
                    //alertUserNoConnection();
                    mmSocket.close();
                    socketConnected = false;
                    //alertUserNoConnection();
                    mHandler.removeCallbacksAndMessages(null);
                    mHandler.obtainMessage(MESSAGE_CONNECT_FAILED).sendToTarget();
                    //mHandler.sendEmptyMessage(0);
                } catch (IOException closeException) {
                    Log.e("TAG", "Could not close the client socket");
                }
                return;
            }
            socketConnected = true;
            // The connection attempt succeeded. Perform work associated with
            // the connection in a separate thread.
            //Log.e("TAG","Before Wait");
            //SystemClock.sleep((long)3000);
            //Log.e("TAG", "After Wait");
            manageMyConnectedSocket(mmSocket);
            //mHandler.obtainMessage(SUCCESS_CONNECT, mmSocket).sendToTarget();
        }


        // Closes the client socket and causes the thread to finish.
        public void cancel() {
            try {
                mmSocket.close();
            } catch (IOException e) {}
        }
    }

    //***************************************************************************************************//
    //Tried to get TextView to change in the case that no connection was formed, but it doesn't seem
    //to like it when I try to manipulate the TextView from a thread.
    //***************************************************************************************************//
    private void alertUserNoConnection() {
        //TextView textView = (TextView)findViewById(R.id.tvConnected);
        //textView.setText((String)"Socket failed...try again");
    }


    //***************************************************************************************************//
    //This is called from within the ConnectThread, this method creates the ConnectedThread and uses
    //it to read in the input from the C# application.  Depending on what is sent, it will alter volume
    //or make a phone call.  Attempted to get it to write to the C# program to alert it when a call
    //has ended, but this has been unsuccessful
    //***************************************************************************************************//
    @TargetApi(23)
    private void manageMyConnectedSocket(BluetoothSocket mmSocket) {
        Log.e("TAG", "in manageMyConnectedSocket()");
        ConnectedThread connectedThread = new ConnectedThread((BluetoothSocket)mmSocket);
        Log.e("TAG", "connectedThread successfully created");
        isConnected = true;
        String ss = " successfully connected";

        String Volumeup = "VOLUME_UP";
        String Volumedown = "VOLUME_DOWN";
        while(mmSocket.isConnected())
        {
            if(currentState.equals(STATE_CALL_ENDED))
            {
                currentState = STATE_FIRST_START;
                connectedThread.write("Call ended...".getBytes());

            }
            if (readFail < 10) {
                ss = connectedThread.read();
            }
            else
            {
                Log.e("TAG", "Read() failed too many times...Connection possibly lost");
                mHandler.removeCallbacksAndMessages(null);
                mHandler.obtainMessage(MESSAGE_CONNECT_FAILED).sendToTarget();
                //TextView textView = (TextView)findViewById(R.id.tvConnected);
                //textView.setText((String)"Socket failed...try again");
                break;
            }


            if(ss.equals(Volumeup))
            {
                Log.e("TAG", "In Volume_up");
                AudioManager audioManager = (AudioManager)getSystemService(Context.AUDIO_SERVICE);
                int newVolume = audioManager.getStreamVolume(AudioManager.STREAM_VOICE_CALL);
                if (newVolume <= audioManager.getStreamMaxVolume(AudioManager.STREAM_VOICE_CALL))
                {
                    newVolume++;
                }
                audioManager.setStreamVolume(AudioManager.STREAM_VOICE_CALL, newVolume, AudioManager.FLAG_SHOW_UI);
            }
            else if(ss.equals(Volumedown))
            {
                Log.e("TAG", "In Volume_down");
                AudioManager audioManager = (AudioManager)getSystemService(Context.AUDIO_SERVICE);
                int newVolume = audioManager.getStreamVolume(AudioManager.STREAM_VOICE_CALL);
                if (newVolume >= 0)
                {
                    newVolume--;
                }
                audioManager.setStreamVolume(AudioManager.STREAM_VOICE_CALL, newVolume, AudioManager.FLAG_SHOW_UI);
            }

            else if (!ss.equals("FAILED"))        //CALL
            {
                Intent callIntent = new Intent(Intent.ACTION_CALL);
                callIntent.setData(Uri.parse("tel:" + ss));

                if (ActivityCompat.checkSelfPermission(MainActivity.this,
                        Manifest.permission.CALL_PHONE) != PackageManager.PERMISSION_GRANTED) {
                    return;
                }

                startActivity(callIntent);

                currentState = STATE_CALL_MADE;


                //Uses Call class, currently unable to create an object, so this is being put on hold...
                //Log.e("TAG", "Before trying Call class");
/*
                while (call.getState() != Call.STATE_ACTIVE || call.getState() != Call.STATE_DISCONNECTED)
                {
                    //wait
                }
                Log.e("TAG", "After while call.getState()");

                ss = "declined";
                if(call.getState() == Call.STATE_ACTIVE) {
                    ss = "answered";
                }

                connectedThread.write(ss.getBytes());
*/

            }


        }
        //Toast.makeText(getApplicationContext(), "Connection was ended...", Toast.LENGTH_SHORT).show();
        Log.e("TAG", "Connection was ended...");
    }

    // Bluetooth Connected Thread from developer.android.com
    private class ConnectedThread extends Thread {
        private final BluetoothSocket mmSocket;
        private final InputStream mmInStream;
        private final OutputStream mmOutStream;
        private byte[] mmBuffer; // mmBuffer store for the stream

        public ConnectedThread(BluetoothSocket socket) {
            mmSocket = socket;
            InputStream tmpIn = null;
            OutputStream tmpOut = null;

            // Get the input and output streams; using temp objects because
            // member streams are final.
            try {
                tmpIn = socket.getInputStream();
            } catch (IOException e) {}
            try {
                tmpOut = socket.getOutputStream();
            } catch (IOException e) {}

            mmInStream = tmpIn;
            mmOutStream = tmpOut;
        }

        public void run() {
            //mmBuffer = new byte[1024];    //moved into loop so the buffer will be new every time
            int numBytes; // bytes returned from read()

            // Keep listening to the InputStream until an exception occurs.
            while (true) {
                try {
                    // Read from the InputStream.
                    mmBuffer = new byte[1024];
                    numBytes = mmInStream.read(mmBuffer);
                    // Send the obtained bytes to the UI activity.
                    mHandler.obtainMessage(MESSAGE_READ, numBytes, -1, mmBuffer).sendToTarget();

                } catch (IOException e) {
                    break;
                }
            }
        }

        //***************************************************************************************************//
        //It looks like the run() function handles reads for this thread, but I didn't want to loop inside
        //thread call, so to make it simpler for myself I made this read function that calls my own
        //read method where I return the message to here
        //***************************************************************************************************//
        public String read() {
            try {
                mmBuffer = new byte[1024];
                int numBytes = mmInStream.read(mmBuffer);
                Log.e("TAG", "Right before readMessage");
                //mHandler.obtainMessage(MESSAGE_READ, numBytes, -1, mmBuffer).sendToTarget();

                return readMessage(mmBuffer, numBytes);

            }   catch(IOException e) {
                Log.e("TAG", "read() failed...");
                readFail++;
                return "FAILED";
            }
        }

        // Call this from the main activity to send data to the remote device.
        public void write(byte[] bytes) {
            try {
                mmOutStream.write(bytes);

                // Share the sent message with the UI activity.
                /*mHandler.obtainMessage(MESSAGE_WRITE, -1, -1, mmBuffer)
                        .sendToTarget();*/
            } catch (IOException e) {

                // Send a failure message back to the activity.
                /*Message writeErrorMsg =
                        mHandler.obtainMessage(MessageConstants.MESSAGE_TOAST);
                Bundle bundle = new Bundle();
                bundle.putString("toast",
                        "Couldn't send data to the other device");
                writeErrorMsg.setData(bundle);
                mHandler.sendMessage(writeErrorMsg);*/
            }
        }

        // Call this method from the main activity to shut down the connection.
        public void cancel() {
            try {
                mmSocket.close();
            } catch (IOException e) {
            }
        }
    }

    //***************************************************************************************************//
    //Reads in the message stored in mmBuffer and changes it to a string to return to the thread handler
    //***************************************************************************************************//
    @TargetApi(23)
    private String readMessage(byte[] mmBuffer, int numBytes) {
        Log.e("TAG","In readMessage");
        byte[] readBuf = new byte[numBytes];
        for (int i=0; i < readBuf.length; i++)
        {
            readBuf[i] = mmBuffer[i];
        }
        String ss = new String(readBuf);
        Log.e("TAG", "Right after mmBuffer used");
        Log.e("TAG", ss + ";");
        return ss;

        /*
        String Volumeup = "VOLUME_UP";
        String Volumedown = "VOLUME_DOWN";
        if(ss.equals(Volumeup))
        {
            Log.e("TAG", "In Volume_up");
            AudioManager audioManager = (AudioManager)getSystemService(Context.AUDIO_SERVICE);
            int newVolume = audioManager.getStreamVolume(AudioManager.STREAM_VOICE_CALL);
            if (newVolume <= audioManager.getStreamMaxVolume(AudioManager.STREAM_VOICE_CALL))
            {
                newVolume++;
            }
            audioManager.setStreamVolume(AudioManager.STREAM_VOICE_CALL, newVolume, AudioManager.FLAG_SHOW_UI);
        }
        else if(ss.equals(Volumedown))
        {
            Log.e("TAG", "In Volume_down");
            AudioManager audioManager = (AudioManager)getSystemService(Context.AUDIO_SERVICE);
            int newVolume = audioManager.getStreamVolume(AudioManager.STREAM_VOICE_CALL);
            if (newVolume >= 0)
            {
                newVolume--;
            }
            audioManager.setStreamVolume(AudioManager.STREAM_VOICE_CALL, newVolume, AudioManager.FLAG_SHOW_UI);
        }
        else        //CALL
        {
            Intent callIntent = new Intent(Intent.ACTION_CALL);
            callIntent.setData(Uri.parse("tel:" + ss));

            if (ActivityCompat.checkSelfPermission(MainActivity.this,
                    Manifest.permission.CALL_PHONE) != PackageManager.PERMISSION_GRANTED) {
                return;
            }

            startActivity(callIntent);

            Log.e("TAG", "Before trying Call class");
            Call call = null;
            while (call.getState() != Call.STATE_ACTIVE)
            {
                //wait
            }
            Log.e("TAG", "After while call.getState()");




        }*/

    }
}